from __future__ import annotations

import json
from typing import Any

from tracelm.trace import Trace


def export_trace_to_chrome(trace: Trace, output_file: str) -> None:
    root = trace.get_root_span()
    if root is not None:
        baseline = float(root.start_time)
    elif trace.spans:
        baseline = min(float(span.start_time) for span in trace.spans.values())
    else:
        baseline = 0.0

    events: list[dict[str, Any]] = []
    for span in trace.spans.values():
        ts = int((float(span.start_time) - baseline) * 1_000_000)
        dur = int(float(span.duration) * 1_000_000)
        if ts < 0:
            ts = 0
        if dur < 0:
            dur = 0

        events.append(
            {
                "name": span.name,
                "ph": "X",
                "ts": ts,
                "dur": dur,
                "pid": 1,
                "tid": 1,
            }
        )

    payload = {"traceEvents": events}
    with open(output_file, "w", encoding="utf-8") as f:
        json.dump(payload, f)



from __future__ import annotations

import random


def should_sample(rate: float) -> bool:
    if not 0.0 <= rate <= 1.0:
        raise ValueError("rate must be between 0.0 and 1.0")
    if rate == 1.0:
        return True
    if rate == 0.0:
        return False
    return random.random() < rate



from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Any

from tracelm.cli.tree import render_trace_tree
from tracelm.context import create_new_trace, generate_span_id, get_current_trace, set_current_span
from tracelm.decorator import get_trace
from tracelm.exporters.chrome_exporter import export_trace_to_chrome
from tracelm.exporters.otel_exporter import export_trace_to_otel
from tracelm.profiler import generate_summary
from tracelm.sampling import should_sample
from tracelm.span import Span
from tracelm.storage.sqlite_store import init_db, list_traces, load_trace, save_trace
from tracelm.trace import Trace


def _resolve_trace_object() -> Trace | None:
    current = get_current_trace()
    if isinstance(current, Trace):
        return current
    if isinstance(current, str):
        return get_trace(current)
    return None


def _trace_from_data(data: dict[str, Any]) -> Trace:
    trace = Trace(trace_id=str(data.get("trace_id", "")))
    spans_data = data.get("spans", {})

    if isinstance(spans_data, dict):
        for span_id, span_payload in spans_data.items():
            if not isinstance(span_payload, dict):
                continue
            payload = dict(span_payload)
            payload.setdefault("span_id", span_id)
            payload.setdefault("trace_id", trace.trace_id)
            payload.setdefault("name", str(payload.get("span_id", span_id)))
            span = Span(**payload)
            trace.spans[span.span_id] = span

    root_span_id = data.get("root_span_id")
    trace.root_span_id = root_span_id if isinstance(root_span_id, str) else None
    return trace


def _cmd_run(python_file: str, otel: bool = False, sample_rate: float = 1.0) -> None:
    source = Path(python_file).read_text(encoding="utf-8")

    if not should_sample(sample_rate):
        from tracelm.context import set_tracing_enabled

        set_tracing_enabled(False)
        exec(open(python_file).read(), {})
        return

    create_new_trace()

    from tracelm import decorator as _decorator

    trace_ref = get_current_trace()
    if isinstance(trace_ref, Trace):
        trace = trace_ref
    elif isinstance(trace_ref, str):
        trace = get_trace(trace_ref)
        if trace is None:
            trace = Trace(trace_id=trace_ref)
            _decorator._TRACE_REGISTRY[trace_ref] = trace
    else:
        raise RuntimeError("No active trace. Use CLI run command.")

    root_span = Span(
        span_id=generate_span_id(),
        trace_id=trace.trace_id,
        parent_id=None,
        name="__root__",
    )

    trace.add_span(root_span)
    set_current_span(root_span)

    exec(source, {})

    from tracelm.context import get_current_span

    current_span = get_current_span()
    if current_span is not None:
        current_span.finish()

    trace = _resolve_trace_object()
    if trace is not None:
        trace.validate()
        if otel:
            from tracelm.bridges.otel_bridge import enable_otel_bridge, export_trace_to_otel_sdk

            enable_otel_bridge()
            export_trace_to_otel_sdk(trace)
        save_trace(trace)
        summary = generate_summary(trace)
        critical_path = " -> ".join(summary["critical_path"])

        print("Trace Summary")
        print("-------------")
        print(f"Trace ID: {summary['trace_id']}")
        print(f"Total Latency: {summary['total_latency']}")
        print(f"Total Spans: {summary['total_spans']}")
        print(f"Slowest Span: {summary['slowest_span']}")
        print(f"Critical Path: {critical_path}")
        print(f"Tokens In: {summary['total_tokens_in']}")
        print(f"Tokens Out: {summary['total_tokens_out']}")
        print(f"Total Cost: {summary['total_cost']}")
        print(f"Anomalies: {summary['anomalies']}")
        print("")
        print("Execution Tree")
        print("--------------")
        print(render_trace_tree(trace))


def _cmd_analyze(trace_id: str) -> None:
    data = load_trace(trace_id)
    if data is None:
        print("null")
        return
    print(json.dumps(data))


def _cmd_export(trace_id: str, export_format: str) -> None:
    data = load_trace(trace_id)
    if data is None:
        print("null")
        return

    trace = _trace_from_data(data)

    if export_format == "chrome":
        output_file = f"trace_{trace_id}.json"
        export_trace_to_chrome(trace, output_file)
        print(f"Exported to {output_file}")
        return

    if export_format == "otel":
        output_file = f"trace_{trace_id}_otel.json"
        export_trace_to_otel(trace, output_file)
        print(f"Exported to {output_file}")


def _cmd_compare(trace_id_1: str, trace_id_2: str) -> None:
    data_1 = load_trace(trace_id_1)
    data_2 = load_trace(trace_id_2)

    if data_1 is None or data_2 is None:
        print("null")
        return

    trace_1 = _trace_from_data(data_1)
    trace_2 = _trace_from_data(data_2)

    summary_1 = generate_summary(trace_1)
    summary_2 = generate_summary(trace_2)

    latency_1 = summary_1["total_latency"]
    latency_2 = summary_2["total_latency"]
    cost_1 = summary_1["total_cost"]
    cost_2 = summary_2["total_cost"]
    tokens_out_1 = summary_1["total_tokens_out"]
    tokens_out_2 = summary_2["total_tokens_out"]

    print("Trace Comparison")
    print("----------------")
    print(f"Latency 1: {latency_1}")
    print(f"Latency 2: {latency_2}")
    print(f"Latency Delta: {latency_2 - latency_1}")
    print("")
    print(f"Cost 1: {cost_1}")
    print(f"Cost 2: {cost_2}")
    print(f"Cost Delta: {cost_2 - cost_1}")
    print("")
    print(f"Token Delta (out): {tokens_out_2 - tokens_out_1}")


def _cmd_list() -> None:
    for trace_id in list_traces():
        print(trace_id)


def run(argv: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(prog="tracelm")
    subparsers = parser.add_subparsers(dest="command", required=True)

    run_parser = subparsers.add_parser("run")
    run_parser.add_argument("python_file")
    run_parser.add_argument("--otel", action="store_true")
    run_parser.add_argument("--sample-rate", type=float, default=1.0)

    analyze_parser = subparsers.add_parser("analyze")
    analyze_parser.add_argument("trace_id")

    export_parser = subparsers.add_parser("export")
    export_parser.add_argument("trace_id")
    export_parser.add_argument("--format", dest="export_format", choices=["chrome", "otel"], required=True)

    compare_parser = subparsers.add_parser("compare")
    compare_parser.add_argument("trace_id_1")
    compare_parser.add_argument("trace_id_2")

    subparsers.add_parser("list")

    args = parser.parse_args(argv)
    init_db()

    if args.command == "run":
        _cmd_run(args.python_file, otel=args.otel, sample_rate=args.sample_rate)
        return
    if args.command == "analyze":
        _cmd_analyze(args.trace_id)
        return
    if args.command == "export":
        _cmd_export(args.trace_id, args.export_format)
        return
    if args.command == "compare":
        _cmd_compare(args.trace_id_1, args.trace_id_2)
        return
    if args.command == "list":
        _cmd_list()


if __name__ == "__main__":
    run()



from __future__ import annotations

from typing import Any, Callable

from tracelm.context import get_current_span, get_current_trace
from tracelm.distributed.tracecontext import build_traceparent

_original_request: Callable[..., Any] | None = None
_is_patched = False


def patch_requests() -> None:
    global _original_request, _is_patched

    if _is_patched:
        return

    import requests

    _original_request = requests.Session.request

    def wrapper(self: requests.Session, method: str, url: str, **kwargs: Any) -> Any:
        trace = get_current_trace()
        span = get_current_span()

        trace_id: str | None = None
        if isinstance(trace, str):
            trace_id = trace
        elif trace is not None:
            trace_id = getattr(trace, "trace_id", None)

        span_id = getattr(span, "span_id", None)

        if trace_id and span_id:
            headers = kwargs.get("headers")
            if headers is None:
                headers = {}
                kwargs["headers"] = headers

            if "traceparent" not in headers:
                headers["traceparent"] = build_traceparent(trace_id, span_id)

        return _original_request(self, method, url, **kwargs)

    requests.Session.request = wrapper
    _is_patched = True



from __future__ import annotations


def _is_hex(value: str) -> bool:
    if not value:
        return False
    return all(ch in "0123456789abcdefABCDEF" for ch in value)


def parse_traceparent(header: str) -> tuple[str, str] | None:
    parts = header.split("-")
    if len(parts) != 4:
        return None

    _, trace_id, span_id, _ = parts

    if len(trace_id) != 32:
        return None
    if len(span_id) != 16:
        return None
    if not _is_hex(trace_id) or not _is_hex(span_id):
        return None

    return trace_id.lower(), span_id.lower()


def build_traceparent(trace_id: str, span_id: str) -> str:
    return f"00-{trace_id}-{span_id}-01"



"""Execution context helpers for trace and span tracking.

This module uses ``contextvars`` so trace/span state is isolated per execution
context (thread, task, coroutine) and does not leak across concurrent work.
"""

from __future__ import annotations

import secrets
from contextvars import ContextVar, Token
from typing import Any

_current_trace: ContextVar[str | None] = ContextVar("current_trace", default=None)
_current_span: ContextVar[Any | None] = ContextVar("current_span", default=None)
_tracing_enabled: bool = True


def set_tracing_enabled(enabled: bool) -> None:
    global _tracing_enabled
    _tracing_enabled = enabled


def is_tracing_enabled() -> bool:
    return _tracing_enabled


def generate_trace_id() -> str:
    while True:
        trace_id = secrets.token_hex(16).lower()
        if trace_id != "0" * 32:
            return trace_id


def generate_span_id() -> str:
    while True:
        span_id = secrets.token_hex(8).lower()
        if span_id != "0" * 16:
            return span_id


def create_new_trace() -> str:
    trace_id = generate_trace_id()
    _current_trace.set(trace_id)
    _current_span.set(None)
    return trace_id


def get_current_trace() -> str | None:
    return _current_trace.get()


def set_current_span(span: Any | None) -> Token[Any | None]:
    return _current_span.set(span)


def get_current_span() -> Any | None:
    return _current_span.get()


def record_tokens(tokens_in: int = 0, tokens_out: int = 0, cost: float = 0.0) -> None:
    span = get_current_span()
    if span is None:
        return

    try:
        span.tokens_in += tokens_in
        span.tokens_out += tokens_out
        span.cost += cost
    except Exception:
        return



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional, Set

from tracelm.span import Span


@dataclass(slots=True)
class Trace:
    trace_id: str
    spans: Dict[str, Span] = field(default_factory=dict)
    root_span_id: Optional[str] = None

    def add_span(self, span: Span) -> None:
        if span.trace_id != self.trace_id:
            raise ValueError("span.trace_id must match trace.trace_id")

        self.spans[span.span_id] = span

        if span.parent_id is None:
            self.root_span_id = span.span_id
        else:
            parent = self.spans.get(span.parent_id)
            if parent is not None and span.span_id not in parent.children:
                parent.children.append(span.span_id)

        for candidate in self.spans.values():
            if candidate.parent_id == span.span_id and candidate.span_id not in span.children:
                span.children.append(candidate.span_id)

    def get_root_span(self) -> Optional[Span]:
        if self.root_span_id is None:
            return None
        return self.spans.get(self.root_span_id)

    def get_span(self, span_id: str) -> Optional[Span]:
        return self.spans.get(span_id)

    def validate(self) -> None:
        roots = [span for span in self.spans.values() if span.parent_id is None]
        external_parents = [
            span for span in self.spans.values() if span.parent_id is not None and span.parent_id not in self.spans
        ]

        is_local = len(roots) == 1 and len(external_parents) == 0
        is_distributed = len(roots) == 0 and len(external_parents) == 1
        if not (is_local or is_distributed):
            raise ValueError("invalid trace structure")

        entry_span = roots[0] if is_local else external_parents[0]

        children_by_parent: Dict[str, Set[str]] = {span_id: set() for span_id in self.spans}
        for span in self.spans.values():
            if span.parent_id is None or span.parent_id not in self.spans:
                continue
            children_by_parent[span.parent_id].add(span.span_id)

        state: Dict[str, int] = {span_id: 0 for span_id in self.spans}
        visited: Set[str] = set()

        def dfs(span_id: str) -> None:
            status = state[span_id]
            if status == 1:
                raise ValueError("invalid trace structure")
            if status == 2:
                return

            state[span_id] = 1
            visited.add(span_id)
            for child_id in children_by_parent.get(span_id, set()):
                dfs(child_id)
            state[span_id] = 2

        dfs(entry_span.span_id)

        if len(visited) != len(self.spans):
            raise ValueError("invalid trace structure")



from __future__ import annotations

import json
import sqlite3
import time
from typing import Any, Dict, List, Optional

from tracelm.trace import Trace

DB_FILE = "tracelm_traces.db"


def init_db() -> None:
    with sqlite3.connect(DB_FILE) as conn:
        conn.execute(
            """
            CREATE TABLE IF NOT EXISTS traces(
                trace_id TEXT PRIMARY KEY,
                timestamp REAL,
                data TEXT
            )
            """
        )
        conn.commit()


def _span_to_json_dict(span: Any) -> Dict[str, Any]:
    try:
        return json.loads(json.dumps(vars(span)))
    except TypeError:
        fallback = {slot: getattr(span, slot) for slot in getattr(span, "__slots__", [])}
        return json.loads(json.dumps(fallback))


def save_trace(trace: Trace) -> None:
    init_db()
    spans_payload = {span_id: _span_to_json_dict(span) for span_id, span in trace.spans.items()}
    payload = {
        "trace_id": trace.trace_id,
        "root_span_id": trace.root_span_id,
        "spans": spans_payload,
    }

    with sqlite3.connect(DB_FILE) as conn:
        conn.execute(
            """
            INSERT OR REPLACE INTO traces(trace_id, timestamp, data)
            VALUES(?, ?, ?)
            """,
            (trace.trace_id, time.time(), json.dumps(payload)),
        )
        conn.commit()


def load_trace(trace_id: str) -> dict | None:
    init_db()
    with sqlite3.connect(DB_FILE) as conn:
        row = conn.execute("SELECT data FROM traces WHERE trace_id = ?", (trace_id,)).fetchone()
    if row is None:
        return None
    return json.loads(row[0])


def list_traces() -> List[str]:
    init_db()
    with sqlite3.connect(DB_FILE) as conn:
        rows = conn.execute("SELECT trace_id FROM traces ORDER BY timestamp DESC").fetchall()
    return [str(row[0]) for row in rows]



from __future__ import annotations

from dataclasses import dataclass, field
from time import perf_counter
from typing import Any, Dict, List, Optional

from tracelm.context import generate_span_id


def _is_lower_hex(value: str, length: int) -> bool:
    if len(value) != length:
        return False
    allowed = set("0123456789abcdef")
    return all(ch in allowed for ch in value)


@dataclass(slots=True)
class Span:
    trace_id: str
    name: str
    span_id: str = field(default_factory=generate_span_id)
    parent_id: Optional[str] = None
    start_time: float = field(default_factory=perf_counter)
    end_time: float = 0.0
    duration: float = 0.0
    tokens_in: int = 0
    tokens_out: int = 0
    cost: float = 0.0
    error: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    children: List[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.span_id = self.span_id.lower()
        self.trace_id = self.trace_id.lower()
        if self.parent_id is not None:
            self.parent_id = self.parent_id.lower()

        if not self.span_id:
            raise ValueError("span_id must be a non-empty string")
        if not _is_lower_hex(self.span_id, 16) or self.span_id == "0" * 16:
            raise ValueError("span_id must be 16 lowercase hex characters and not all zeros")

        if not self.trace_id:
            raise ValueError("trace_id must be a non-empty string")
        if not _is_lower_hex(self.trace_id, 32) or self.trace_id == "0" * 32:
            raise ValueError("trace_id must be 32 lowercase hex characters and not all zeros")

        if self.parent_id is not None:
            if not _is_lower_hex(self.parent_id, 16) or self.parent_id == "0" * 16:
                raise ValueError("parent_id must be 16 lowercase hex characters and not all zeros")

        if not self.name:
            raise ValueError("name must be a non-empty string")
        if self.tokens_in < 0:
            raise ValueError("tokens_in must be >= 0")
        if self.tokens_out < 0:
            raise ValueError("tokens_out must be >= 0")
        if self.cost < 0:
            raise ValueError("cost must be >= 0")

    def finish(self) -> None:
        end = perf_counter()
        if end < self.start_time:
            end = self.start_time
        self.end_time = end
        self.duration = self.end_time - self.start_time



from __future__ import annotations

import sys
from typing import Dict, List, Optional

from tracelm.span import Span
from tracelm.trace import Trace


def _format_span_line(span: Span) -> str:
    duration_ms = float(span.duration) * 1000.0
    return f"{span.name} ({duration_ms:.3f} ms)"


def _find_entry_span(trace: Trace, children_by_parent: Dict[str, List[str]]) -> Optional[Span]:
    roots = [span for span in trace.spans.values() if span.parent_id is None]
    if len(roots) == 1:
        return roots[0]

    distributed_entries = [
        span for span in trace.spans.values() if span.parent_id is not None and span.parent_id not in trace.spans
    ]
    if len(distributed_entries) == 1:
        return distributed_entries[0]

    if trace.spans:
        return min(trace.spans.values(), key=lambda s: s.start_time)

    return None


def render_trace_tree(trace: Trace) -> str:
    if not trace.spans:
        return ""

    encoding = sys.stdout.encoding
    if encoding is None or "utf" not in encoding.lower():
        branch = "+-- "
        last_branch = "\\-- "
        vertical = "|   "
    else:
        branch = "\u251c\u2500\u2500 "
        last_branch = "\u2514\u2500\u2500 "
        vertical = "\u2502   "

    space = "    "

    children_by_parent: Dict[str, List[str]] = {span_id: [] for span_id in trace.spans}
    for span in trace.spans.values():
        if span.parent_id is not None and span.parent_id in trace.spans:
            children_by_parent[span.parent_id].append(span.span_id)

    for parent_id, child_ids in children_by_parent.items():
        child_ids.sort(key=lambda child_id: trace.spans[child_id].start_time)

    entry = _find_entry_span(trace, children_by_parent)
    if entry is None:
        return ""

    lines: List[str] = [_format_span_line(entry)]

    def walk(span_id: str, prefix: str) -> None:
        child_ids = children_by_parent.get(span_id, [])
        for index, child_id in enumerate(child_ids):
            child = trace.spans[child_id]
            is_last = index == len(child_ids) - 1
            connector = last_branch if is_last else branch
            lines.append(f"{prefix}{connector}{_format_span_line(child)}")
            next_prefix = prefix + (space if is_last else vertical)
            walk(child_id, next_prefix)

    walk(entry.span_id, "")
    return "\n".join(lines)



from __future__ import annotations

from typing import Any

from tracelm.span import Span
from tracelm.trace import Trace

_tracer: Any | None = None


def enable_otel_bridge(service_name: str = "tracelm") -> None:
    global _tracer

    try:
        from opentelemetry import trace
        from opentelemetry.sdk.resources import Resource
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor
    except ImportError as exc:
        raise ImportError(
            "OpenTelemetry bridge requires optional dependency 'opentelemetry-sdk'. "
            "Install with: pip install opentelemetry-sdk"
        ) from exc

    resource = Resource.create({"service.name": service_name})
    provider = TracerProvider(resource=resource)
    processor = SimpleSpanProcessor(ConsoleSpanExporter())
    provider.add_span_processor(processor)
    trace.set_tracer_provider(provider)
    _tracer = trace.get_tracer("tracelm")


def create_otel_span(span: Span) -> None:
    if _tracer is None:
        raise RuntimeError("OpenTelemetry bridge is not enabled. Call enable_otel_bridge() first.")

    with _tracer.start_as_current_span(span.name) as otel_span:
        otel_span.set_attribute("tokens.in", int(span.tokens_in))
        otel_span.set_attribute("tokens.out", int(span.tokens_out))
        otel_span.set_attribute("cost", float(span.cost))


def export_trace_to_otel_sdk(trace: Trace) -> None:
    for span in trace.spans.values():
        create_otel_span(span)



from __future__ import annotations

import json
from typing import Any

from tracelm.trace import Trace


def _normalize_hex(value: str) -> str:
    return value.replace("-", "")


def export_trace_to_otel(trace: Trace, output_file: str) -> None:
    trace_id = _normalize_hex(trace.trace_id)

    spans: list[dict[str, Any]] = []
    for span in trace.spans.values():
        span_id = _normalize_hex(span.span_id)[:16]
        parent_span_id = ""
        if span.parent_id:
            parent_span_id = _normalize_hex(span.parent_id)[:16]

        spans.append(
            {
                "traceId": trace_id,
                "spanId": span_id,
                "parentSpanId": parent_span_id,
                "name": span.name,
                "startTimeUnixNano": int(float(span.start_time) * 1_000_000_000),
                "endTimeUnixNano": int(float(span.end_time) * 1_000_000_000),
                "attributes": [
                    {"key": "tokens.in", "value": {"intValue": int(span.tokens_in)}},
                    {"key": "tokens.out", "value": {"intValue": int(span.tokens_out)}},
                    {"key": "cost", "value": {"doubleValue": float(span.cost)}},
                ],
            }
        )

    payload = {
        "resourceSpans": [
            {
                "resource": {
                    "attributes": [
                        {"key": "service.name", "value": {"stringValue": "tracelm"}},
                    ]
                },
                "scopeSpans": [
                    {
                        "scope": {"name": "tracelm"},
                        "spans": spans,
                    }
                ],
            }
        ]
    }

    with open(output_file, "w", encoding="utf-8") as f:
        json.dump(payload, f)



from __future__ import annotations

from functools import wraps
from inspect import iscoroutinefunction
from typing import Any, Callable, Dict, Optional, TypeVar, cast

from tracelm.context import (
    generate_span_id,
    get_current_span,
    get_current_trace,
    is_tracing_enabled,
    set_current_span,
)
from tracelm.span import Span
from tracelm.trace import Trace

F = TypeVar("F", bound=Callable[..., Any])

_TRACE_REGISTRY: Dict[str, Trace] = {}


def get_trace(trace_id: str) -> Optional[Trace]:
    return _TRACE_REGISTRY.get(trace_id)


def _get_active_trace() -> Trace:
    current_trace = get_current_trace()

    if isinstance(current_trace, Trace):
        _TRACE_REGISTRY[current_trace.trace_id] = current_trace
        return current_trace

    if isinstance(current_trace, str):
        existing = _TRACE_REGISTRY.get(current_trace)
        if existing is not None:
            return existing
        new_trace = Trace(trace_id=current_trace)
        _TRACE_REGISTRY[current_trace] = new_trace
        return new_trace

    raise RuntimeError("No active trace. Use CLI run command.")


def node(name: str) -> Callable[[F], F]:
    if not name:
        raise ValueError("name must be a non-empty string")

    def decorator(func: F) -> F:
        if iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                if not is_tracing_enabled():
                    return await func(*args, **kwargs)

                trace = _get_active_trace()
                parent_span = get_current_span()
                if not isinstance(parent_span, Span):
                    raise RuntimeError("No active parent span. CLI must initialize root span.")
                parent_id = parent_span.span_id

                span = Span(
                    span_id=generate_span_id(),
                    trace_id=trace.trace_id,
                    parent_id=parent_id,
                    name=name,
                )
                trace.add_span(span)
                set_current_span(span)

                try:
                    return await func(*args, **kwargs)
                except Exception as exc:
                    span.error = f"{type(exc).__name__}: {exc}"
                    raise
                finally:
                    span.finish()
                    set_current_span(parent_span)

            return cast(F, async_wrapper)

        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            if not is_tracing_enabled():
                return func(*args, **kwargs)

            trace = _get_active_trace()
            parent_span = get_current_span()
            if not isinstance(parent_span, Span):
                raise RuntimeError("No active parent span. CLI must initialize root span.")
            parent_id = parent_span.span_id

            span = Span(
                span_id=generate_span_id(),
                trace_id=trace.trace_id,
                parent_id=parent_id,
                name=name,
            )
            trace.add_span(span)
            set_current_span(span)

            try:
                return func(*args, **kwargs)
            except Exception as exc:
                span.error = f"{type(exc).__name__}: {exc}"
                raise
            finally:
                span.finish()
                set_current_span(parent_span)

        return cast(F, sync_wrapper)

    return decorator



from __future__ import annotations

from typing import Dict, List, Tuple

from tracelm.span import Span
from tracelm.trace import Trace


def compute_total_latency(trace: Trace) -> float:
    root = trace.get_root_span()
    if root is None:
        return 0.0
    return float(root.duration)


def find_slowest_span(trace: Trace) -> Span | None:
    user_spans = [span for span in trace.spans.values() if span.name != "__root__"]
    if not user_spans:
        return None
    return max(user_spans, key=lambda span: span.duration)


def compute_critical_path(trace: Trace) -> list[str]:
    root = trace.get_root_span()
    if root is None:
        return []

    children_by_parent: Dict[str, List[str]] = {span_id: [] for span_id in trace.spans}
    for span in trace.spans.values():
        if span.parent_id is not None and span.parent_id in children_by_parent:
            children_by_parent[span.parent_id].append(span.span_id)

    memo: Dict[str, Tuple[float, List[str]]] = {}

    def dfs(span_id: str) -> Tuple[float, List[str]]:
        cached = memo.get(span_id)
        if cached is not None:
            return cached

        span = trace.get_span(span_id)
        if span is None:
            result = (0.0, [])
            memo[span_id] = result
            return result

        best_child_total = 0.0
        best_child_path: List[str] = []

        for child_id in children_by_parent.get(span_id, []):
            child_total, child_path = dfs(child_id)
            if child_total > best_child_total:
                best_child_total = child_total
                best_child_path = child_path

        total = float(span.duration) + best_child_total
        path = [span.name, *best_child_path]
        result = (total, path)
        memo[span_id] = result
        return result

    return dfs(root.span_id)[1]


def detect_anomalies(trace: Trace) -> dict:
    user_spans = [span for span in trace.spans.values() if span.name != "__root__"]
    if not user_spans:
        return {"latency_spikes": []}

    durations = [span.duration for span in user_spans]
    mean_duration = sum(durations) / len(durations)
    threshold = mean_duration * 2

    spikes = [span.name for span in user_spans if span.duration > threshold]
    return {"latency_spikes": spikes}


def generate_summary(trace: Trace) -> dict:
    slowest = find_slowest_span(trace)
    total_tokens_in = sum(span.tokens_in for span in trace.spans.values())
    total_tokens_out = sum(span.tokens_out for span in trace.spans.values())
    total_cost = sum(span.cost for span in trace.spans.values())

    return {
        "trace_id": trace.trace_id,
        "total_latency": compute_total_latency(trace),
        "total_spans": len(trace.spans),
        "slowest_span": slowest.name if slowest is not None else "",
        "critical_path": compute_critical_path(trace),
        "total_tokens_in": total_tokens_in,
        "total_tokens_out": total_tokens_out,
        "total_cost": total_cost,
        "anomalies": detect_anomalies(trace),
    }



from __future__ import annotations

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import Response

from tracelm import context as _context
from tracelm import decorator as _decorator
from tracelm.context import create_new_trace, generate_span_id, get_current_trace, set_current_span
from tracelm.decorator import get_trace
from tracelm.distributed.tracecontext import build_traceparent, parse_traceparent
from tracelm.profiler import generate_summary
from tracelm.sampling import should_sample
from tracelm.span import Span
from tracelm.storage.sqlite_store import save_trace
from tracelm.trace import Trace


class TraceLMMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, store_traces: bool = True, sample_rate: float = 1.0) -> None:
        super().__init__(app)
        self.store_traces = store_traces
        self.sample_rate = sample_rate

    async def dispatch(self, request: Request, call_next) -> Response:
        if not should_sample(self.sample_rate):
            return await call_next(request)

        incoming_parent_id: str | None = None
        traceparent = request.headers.get("traceparent")

        if traceparent:
            parsed = parse_traceparent(traceparent)
            if parsed is not None:
                incoming_trace_id, incoming_parent_id = parsed
                _context._current_trace.set(incoming_trace_id)
                set_current_span(None)
                trace = get_trace(incoming_trace_id)
                if trace is None:
                    trace = Trace(trace_id=incoming_trace_id)
                    _decorator._TRACE_REGISTRY[incoming_trace_id] = trace
            else:
                create_new_trace()
                trace = None
        else:
            create_new_trace()
            trace = None

        if trace is None:
            trace_ref = get_current_trace()
            if isinstance(trace_ref, Trace):
                trace = trace_ref
            elif isinstance(trace_ref, str):
                trace = get_trace(trace_ref)
                if trace is None:
                    trace = Trace(trace_id=trace_ref)
                    _decorator._TRACE_REGISTRY[trace_ref] = trace
            else:
                raise RuntimeError("No active trace. Use CLI run command.")

        root_span = Span(
            span_id=generate_span_id(),
            trace_id=trace.trace_id,
            parent_id=incoming_parent_id,
            name=f"{request.method} {request.url.path}",
        )
        trace.add_span(root_span)
        set_current_span(root_span)

        try:
            response = await call_next(request)
            root_span.finish()
            response.headers["traceparent"] = build_traceparent(trace.trace_id, root_span.span_id)
            trace.validate()
            _ = generate_summary(trace)
            if self.store_traces:
                save_trace(trace)
            return response
        except Exception:
            root_span.finish()
            raise
